@@ -28,6 +28,7 @@
 from base64 import b32encode
 from yubikit.yubiotp import (
     SLOT,
+    NDEF_TYPE,
     YubiOtpSession,
     YubiOtpSlotConfiguration,
     HmacSha1SlotConfiguration,
@@ -36,7 +37,12 @@ from yubikit.yubiotp import (
     UpdateConfiguration,
 )
 from yubikit.core import TRANSPORT, CommandError
-from yubikit.core.otp import modhex_encode, modhex_decode, OtpConnection
+from yubikit.core.otp import (
+    MODHEX_ALPHABET,
+    modhex_encode,
+    modhex_decode,
+    OtpConnection,
+)
 
 from .util import (
     ykman_group,
@@ -105,7 +111,7 @@ def _failed_to_write_msg(ctx, exc_info):
     logger.error("Failed to write to device", exc_info=exc_info)
     cli_fail(
         "Failed to write to the YubiKey. Make sure the device does not "
-        "have restricted access."
+        'have restricted access (see "ykman otp --help" for more info).'
     )
 
 
@@ -125,7 +131,7 @@ def _confirm_slot_overwrite(slot_state, slot):
     "--access-code",
     required=False,
     metavar="HEX",
-    help="A 6 byte access code. Set to empty to use a prompt for input.",
+    help='A 6 byte access code. Use "-" as a value to prompt for input.',
 )
 def otp(ctx, access_code):
     """
@@ -137,7 +143,9 @@ def otp(ctx, access_code):
     A slot configuration may be write-protected with an access code. This
     prevents the configuration to be overwritten without the access code
     provided. Mode switching the YubiKey is not possible when a slot is
-    configured with an access code.
+    configured with an access code. To provide an access code to commands
+    which require it, use the --access-code option. Note that this option must
+    be given directly after the "otp" command, before any sub-command.
 
     Examples:
 
@@ -156,12 +164,16 @@ def otp(ctx, access_code):
     \b
       Program a random 38 characters long static password to slot 2:
       $ ykman otp static --generate 2 --length 38
+
+    \b
+      Remove a currently set access code from slot 2):
+      $ ykman otp --access-code 0123456789ab settings 2 --delete-access-code
     """
 
     ctx.obj["session"] = YubiOtpSession(ctx.obj["conn"])
     if access_code is not None:
-        if access_code == "":
-            access_code = click_prompt("Enter the access code", show_default=False)
+        if access_code == "-":
+            access_code = click_prompt("Enter the access code", hide_input=True)
 
         try:
             access_code = parse_access_code_hex(access_code)
@@ -208,13 +220,22 @@ def swap(ctx):
 @click_slot_argument
 @click.pass_context
 @click.option("-p", "--prefix", help="Added before the NDEF payload. Typically a URI.")
-def ndef(ctx, slot, prefix):
+@click.option(
+    "-t",
+    "--ndef-type",
+    type=EnumChoice(NDEF_TYPE),
+    default="URI",
+    show_default=True,
+    help="NDEF payload type.",
+)
+def ndef(ctx, slot, prefix, ndef_type):
     """
     Configure a slot to be used over NDEF (NFC).
 
-    The default prefix will be used if no prefix is specified:
-
-        "https://my.yubico.com/yk/#"
+    \b
+    If "--prefix" is not specified, a default value will be used, based on the type:
+    - For URI the default value is: "https://my.yubico.com/yk/#"
+    - For TEXT the default is an empty string
     """
     info = ctx.obj["info"]
     session = ctx.obj["session"]
@@ -226,7 +247,7 @@ def ndef(ctx, slot, prefix):
         cli_fail(f"Slot {slot} is empty.")
 
     try:
-        session.set_ndef_configuration(slot, prefix, ctx.obj["access_code"])
+        session.set_ndef_configuration(slot, prefix, ctx.obj["access_code"], ndef_type)
     except CommandError as e:
         _failed_to_write_msg(ctx, e)
 
@@ -365,8 +386,8 @@ def yubiotp(
 
     try:
         public_id = modhex_decode(public_id)
-    except KeyError:
-        ctx.fail("Invalid public ID, must be modhex.")
+    except ValueError:
+        ctx.fail(f"Invalid public ID, must be modhex ({MODHEX_ALPHABET}).")
 
     if not private_id:
         if generate_private_id:
@@ -643,6 +664,17 @@ def calculate(ctx, slot, challenge, totp, digits):
         _failed_to_write_msg(ctx, e)
 
 
+def parse_modhex_or_bcd(value):
+    try:
+        return True, modhex_decode(value)
+    except ValueError:
+        try:
+            int(value)
+            return False, bytes.fromhex(value)
+        except ValueError:
+            raise ValueError("value must be modhex or decimal")
+
+
 @otp.command()
 @click_slot_argument
 @click.argument("key", callback=click_parse_b32_key, required=False)
@@ -654,6 +686,7 @@ def calculate(ctx, slot, challenge, totp, digits):
     help="Number of digits in generated code (default is 6).",
 )
 @click.option("-c", "--counter", type=int, default=0, help="Initial counter value.")
+@click.option("-i", "--identifier", help="Token identifier.")
 @click.option(
     "--no-enter",
     is_flag=True,
@@ -661,11 +694,49 @@ def calculate(ctx, slot, challenge, totp, digits):
 )
 @click_force_option
 @click.pass_context
-def hotp(ctx, slot, key, digits, counter, no_enter, force):
+def hotp(ctx, slot, key, digits, counter, identifier, no_enter, force):
     """
     Program an HMAC-SHA1 OATH-HOTP credential.
+
+    The YubiKey can be configured to output an OATH Token Identifier as a prefix
+    to the OTP itself, which consists of OMP+TT+MUI. Using the "--identifier" option,
+    you may specify the OMP+TT as 4 characters, the MUI as 8 characters, or the full
+    OMP+TT+MUI as 12 characters. If omitted, a default value of "ubhe" will be used for
+    OMP+TT, and the YubiKey serial number will be used as MUI.
     """
     session = ctx.obj["session"]
+
+    mh1 = False
+    mh2 = False
+    if identifier:
+        if identifier == "-":
+            identifier = "ubhe"
+        if len(identifier) == 4:
+            identifier += f"{session.get_serial():08}"
+        elif len(identifier) == 8:
+            identifier = "ubhe" + identifier
+        if len(identifier) != 12:
+            raise ValueError("Incorrect length for token identifier.")
+
+        omp_m, omp = parse_modhex_or_bcd(identifier[:2])
+        tt_m, tt = parse_modhex_or_bcd(identifier[2:4])
+        mui_m, mui = parse_modhex_or_bcd(identifier[4:])
+        if tt_m and not omp_m:
+            raise ValueError("TT can only be modhex encoded if OMP is as well.")
+        if mui_m and not (omp_m and tt_m):
+            raise ValueError(
+                "MUI can only be modhex encoded if OMP and TT are as well."
+            )
+        token_id = omp + tt + mui
+        if mui_m:
+            mh1 = mh2 = True
+        elif tt_m:
+            mh2 = True
+        elif omp_m:
+            mh1 = True
+    else:
+        token_id = b""
+
     if not key:
         while True:
             key = click_prompt("Enter a secret key (base32)")
@@ -683,6 +754,7 @@ def hotp(ctx, slot, key, digits, counter, no_enter, force):
             slot,
             HotpSlotConfiguration(key)
             .imf(counter)
+            .token_id(token_id, mh1, mh2)
             .digits8(int(digits) == 8)
             .append_cr(not no_enter),
             ctx.obj["access_code"],
@@ -749,6 +821,12 @@ def settings(
     if new_access_code and delete_access_code:
         ctx.fail("--new-access-code conflicts with --delete-access-code.")
 
+    if delete_access_code and not ctx.obj["access_code"]:
+        cli_fail(
+            "--delete-access-code used without providing an access code "
+            '(see "ykman otp --help" for more info).'
+        )
+
     if not session.get_config_state().is_configured(slot):
         cli_fail("Not possible to update settings on an empty slot.")
 
@@ -757,7 +835,9 @@ def settings(
             new_access_code = ctx.obj["access_code"]
     else:
         if new_access_code == "-":
-            new_access_code = click_prompt("Enter new access code", show_default=False)
+            new_access_code = click_prompt(
+                "Enter new access code", hide_input=True, confirmation_prompt=True
+            )
 
         try:
             new_access_code = parse_access_code_hex(new_access_code)
